MODULE trsetup

! This module setup table for computation and 
! initializes the plasma profiles

PUBLIC tr_setup
PRIVATE

CONTAINS

  SUBROUTINE tr_setup

    USE trcomm, ONLY: t,tr_nit_allocate
    USE trloop
    IMPLICIT NONE

    CALL tr_setup_table    ! calculate table for matrix generation
    CALL tr_setup_profile  ! calculate initial profile
    CALL tr_nit_allocate

    T=0.D0                 ! time is initialized.
    CALL tr_save_pvprev

    RETURN
  END SUBROUTINE tr_setup

! ***** calculate table for matrix generation *****

  SUBROUTINE tr_setup_table

    USE trcomm, ONLY: ikind,nrmax,nsamax,neqmax,neqrmax,nvrmax,nvmax, &
                      nsa_neq,nva_neq,id_neq,id_neqnr,neq_neqr, &
                      rg,rg_fixed,neqr_neq,tr_neq_allocate,tr_neqr_allocate
    IMPLICIT NONE
    INTEGER(ikind):: neq,nsa,nva,i,neqr,nr

    neqmax=1+3*nsamax
    nvmax=neqmax*(nrmax+1)

    CALL tr_neq_allocate

!   nsa_neq = 0 : magnetic field
!             otherwise : particle species
!   nva_neq = 1 : density
!             2 : toroidal velocity
!             3 : temperature
!   id_neq  = 0 : not solved
!             1 : solved

    neq=1
    nsa_neq(neq)=0
    nva_neq(neq)=1
    id_neq(neq)=0

    DO nsa=1,nsamax
       DO i=1,3
          neq=neq+1
          nsa_neq(neq)=nsa
          nva_neq(neq)=i
          SELECT CASE(i)
          CASE(1)
             id_neq(neq)=0
          CASE(2)
             id_neq(neq)=0
          CASE(3)
             id_neq(neq)=1
          END SELECT
       ENDDO
    ENDDO

    neqr=0
    DO neq=1,neqmax
       SELECT CASE(id_neq(neq))
!         id_neqnr = 0 : fixed to zero
!                    1 : solved normally
!                    2 : fixed to a given value
!                    3 : fixed to a given scale length
       CASE(0) ! equation is not solved in any radius (fixed to zero)
          neqr_neq(neq)=0
          id_neqnr(neq,0:nrmax)=0
       CASE(1,11) ! flat on axis and fixed at plasma surface
          neqr=neqr+1
          neqr_neq(neq)=neqr
          id_neqnr(neq,0:nrmax-1)=1
          id_neqnr(neq,nrmax)=2
       CASE(2,12) ! fixed to zero on axis and fixed at plasma surface
          neqr=neqr+1
          neqr_neq(neq)=neqr
          id_neqnr(neq,0)=2
          id_neqnr(neq,1:nrmax-1)=1
          id_neqnr(neq,nrmax)=2
       CASE(3,13) ! flat on axis and fixed scale length at plasma surface
          neqr=neqr+1
          neqr_neq(neq)=neqr
          id_neqnr(neq,0:nrmax-1)=1
          id_neqnr(neq,nrmax)=3
       CASE(4,14) ! fixed to zero on axis and fixed scale length at surface
          neqr=neqr+1
          neqr_neq(neq)=neqr
          id_neqnr(neq,0)=2
          id_neqnr(neq,1:nrmax-1)=1
          id_neqnr(neq,nrmax)=3
       CASE DEFAULT
          WRITE(6,*) 'XX tr_setup_table: unideficed id_neq:'
          WRITE(6,*) '   neq,id_neq(neq)=',neq,id_neq(neq)
          STOP
       END SELECT

       IF(id_neq(neq) >= 10) THEN  ! fixed for rg >= rg_fixed
          DO nr=1,nrmax
             nsa=nsa_neq(neq)
             nva=nva_neq(neq)
             IF(rg(nr) >= rg_fixed(nva,nsa)) THEN
                id_neqnr(neq,nr)=2
             END IF
          END DO
       END IF
    END DO
    neqrmax=neqr
    nvrmax=neqrmax*(nrmax+1)

    CALL tr_neqr_allocate

    DO neq=1,neqmax
       IF(neqr_neq(neq) /= 0) neq_neqr(neqr)=neq
    ENDDO
          

    RETURN
  END SUBROUTINE tr_setup_table

! ***** calculate inital profile *****

  SUBROUTINE tr_setup_profile

    USE trcomm, ONLY: ikind,rkind,nrmax,nsamax,rg,ra,rn,ru,rt,ns_nsa, &
         pn,pns,pu,pus,ptpr,ptpp,pts,tr_nr_allocate
                      
    IMPLICIT NONE
    INTEGER(ikind):: nr,nsa,ns
    REAL(rkind):: dr,rhol,pt

    CALL tr_nr_allocate

    dr=ra/dble(nrmax)
    DO nr=0,nrmax
       rg(nr)=dble(nr)*dr
       rhol=rg(nr)/ra
       DO nsa=1,nsamax
          ns=ns_nsa(nsa)
          rn(nsa,nr)=(pn(ns)-pns(ns))*(1.D0-rhol**2)+pns(ns)
          ru(nsa,nr)=(pu(ns)-pus(ns))*(1.D0-rhol**2)+pus(ns)
          pt=(ptpr(ns)+2.D0*ptpp(ns))/3.D0
          rt(nsa,nr)=(pt    -pts(ns))*(1.D0-rhol**2)+pts(ns)
       ENDDO
    ENDDO
  END SUBROUTINE tr_setup_profile
END MODULE trsetup
